package Color;

use Exporter;
@ISA = qw(Exporter);
@EXPORT_OK = qw(get_color);


# Value from the interval 0-1 is converted to a color with the colorscale
#   where $min corresponds to $minc, $max corresponds to $maxc, and $avg to $avgc.
#   The colors ($minc,$maxc,$avgc) are strings such as #ffffff.
#   If no $min,$max, and $avg are given, the values 0,1, and 0.5 are used.
#
sub get_color
{
    my ($minc,$avgc,$maxc,$value,$min,$avg,$max) = @_;

    my ($minr,$ming,$minb) = parse_color($minc);
    my ($maxr,$maxg,$maxb) = parse_color($maxc);
    if ( !defined $max ) 
    { 
        $min = 0.0;
        $avg = 0.5;
        $max = 1.0;
    }

    if ( $value<$min ) { $value=$min }
    if ( $value>$max ) { $value=$max }

    $value -= $min;

    my @color = ();
    if ( $avgc eq $minc || $avgc eq $maxc )
    {
        @color = 
            ( 
             $value*($maxr-$minr) + $minr ,
             $value*($maxg-$ming) + $ming ,
             $value*($maxb-$minb) + $minb ,
            );
    }
    else
    {
        my ($avgr,$avgg,$avgb) = parse_color($avgc);
        if ( $value < $avg )
        {
            @color = 
                ( 
                 ($value/$avg)*($avgr-$minr) + $minr ,
                 ($value/$avg)*($avgg-$ming) + $ming ,
                 ($value/$avg)*($avgb-$minb) + $minb ,
                );
        }
        else
        {
            @color = 
                ( 
                 (($value-$avg)/($max-$avg))*($maxr-$avgr) + $avgr ,
                 (($value-$avg)/($max-$avg))*($maxg-$avgg) + $avgg ,
                 (($value-$avg)/($max-$avg))*($maxb-$avgb) + $avgb ,
                );
        }
    }
    if ( $color[0]>255 || $color[0]<0 ) { Exit::exit_nicely("FIXME: The color calculation incorrect: ",join(",",@_)," -> $color[0],$color[1],$color[2]\n") } 
    if ( $color[1]>255 || $color[1]<0 ) { Exit::exit_nicely("FIXME: The color calculation incorrect: ",join(",",@_)," -> $color[0],$color[1],$color[2]\n") } 
    if ( $color[2]>255 || $color[2]<0 ) { Exit::exit_nicely("FIXME: The color calculation incorrect: ",join(",",@_)," -> $color[0],$color[1],$color[2]\n") } 
    return \@color;
}


# Accepts the same arguments as get_color, but returns the color as 
#   a hex string, e.g. #ffffff
sub get_color_string
{
    my $color = get_color(@_);    
    return sprintf "#%02x%02x%02x", $$color[0]*255, $$color[1]*255, $$color[2]*255;
}


# fffff -> 1.0,1.0,1.0
# 00000 ->  .0, .0, .0
sub parse_color
{
    my ($color) = @_;
    if ( $color=~/^#/ ) { $color=$'; }
    my $r = hex(substr($color,0,2)) / 255.;
    my $g = hex(substr($color,2,2)) / 255.;
    my $b = hex(substr($color,4,2)) / 255.;
    return ($r,$g,$b);
}

# 1.0,1.0,1.0 -> ffffff
sub values_to_hexstring
{
    my ($r,$g,$b) = @_;
    return sprintf "%02x%02x%02x",$r*255,$g*255,$b*255;
}

1;
