#include "conf.h"

#include <string.h>
#include <ctype.h>
#include <malloc.h>
#include <stdlib.h>
#include <errno.h>

#define debug(x...) fprintf(stderr,x)
// #define error(x...) fprintf(stderr,x)

#define MAX_LEN 255
char *expand_vars_free(char *val)
{
    char buf[MAX_LEN];
    char *ss, *es, *bs;

    bs = buf;
    ss = val;
    es = val;
    do
    {
        if ( *es=='\\' && *(es+1) && *(es+1)=='#' )
        {
            snprintf(bs,es-ss+1,"%s",ss);
            bs += es-ss;
            ss = es+1;
        }

        es++;
    }
    while (*es);

    snprintf(bs,es-ss+1,"%s",ss);

    free(val);
    return strdup(buf);
}


char *expand_vars_free2(GHashTable *vals, char *val)
{
    char *s, *hold;

    hold = s = val;
    while (s)
    {
        s = strchr(val, '$');
        if ( ! s ) return val;

        char buf[MAX_LEN]; *buf = 0;
        if ( s!=val ) 
        {
            if ( s-val >= MAX_LEN ) 
            {
                error("FIXME: buffer truncated: \"%s\"\n", val);
                return val;
            }
            snprintf(buf, s-val+1, "%s", val);
        }

        s++;
        if ( ! s || *s != '(') return val;
        s++;

        char *e = strchr(s, ')');
        if ( ! e ) return val;
        if ( e == s ) return val;

        *e = 0;
        char *v = g_hash_table_lookup(vals, s);
        if ( ! v )
        {
            error("Warning: undefined variable \"%s\"\n", s);
            return val;
        }

        snprintf(buf+strlen(buf), MAX_LEN-strlen(buf), "%s", v);
        snprintf(buf+strlen(buf), MAX_LEN-strlen(buf), "%s", e+1);

        val  = strdup(buf);
        free(hold);
        hold = s = val;
    }
    error("FIXME: this should not happen: %s\n", val);
    return NULL;
}


GHashTable *conf_read(char *fname)
{
    char *lineptr = NULL;
    size_t len;
    int l=0;
    FILE *fd;
    GHashTable *vals = NULL;

    fd = fopen(fname, "r");
    if ( ! fd )
    {
        error("%s: %s\n", fname, strerror(errno));
        return NULL;
    }

    vals = g_hash_table_new(g_str_hash, g_str_equal);

    while ( getline(&lineptr, &len, fd) > 0 )
    {       
        l++;

        char *p = lineptr;
        while ( p && *p && isspace(*p) ) p++;

        if ( !p || ! *p ) continue;
        if ( *p == '#' ) continue;

        char *skey = p;
        while ( *p && !isspace(*p) ) p++;
        char *key  = strndup(skey,p-skey);

        while ( *p && isspace(*p) ) p++;
        if ( !*p ) 
        { 
            error("could not parse line %d: %s\n", l,lineptr);
            return NULL;
        }

        if ( *p != '=' )
        {
            error("expected '=' on line %d: %s\n", l,lineptr);
            return NULL;
        }

        p++;

        while ( *p && isspace(*p) ) p++;
        if ( !*p )
        {
            error("premature end of line %d: %s\n", l,lineptr);
            return NULL;
        }

        char *sval;
        if ( *p == '"' )
        {
            p++;
            sval = p;
            //while ( *p && (*p!='"' || *(p-1)=='\\') && (*p!='#' || *(p-1)=='\\') ) p++;
            while ( *p )
            {
                if ( *p=='"' && *(p-1)!='\\' ) break;
                p++;
            }
        }
        else
        {
            sval = p;
            while ( *p && !isspace(*p) && (*p!='#' || *(p-1)=='\\') ) p++;
        }

        char *val = strndup(sval,p-sval);
        val = expand_vars_free(val);

        char *v = g_hash_table_lookup(vals, key);
        if ( v )
        {
            int buflen = strlen(v) + strlen(val) + 5;
            char buf[buflen];
            snprintf(buf, buflen, "%s, %s", v, val);
            free(val);
            val = strdup(buf);
        }

        g_hash_table_insert(vals, key, val);
        //debug("conf: %-10s = \"%s\"\n", key,val);
    }

    fclose(fd);
    return vals;
}



void conf_add_string(GHashTable *vals, char *key, char *val)
{
      if (!vals) vals = g_hash_table_new(g_str_hash, g_str_equal);
      g_hash_table_insert(vals, key, val);
}



char *conf_get_string_or_die(GHashTable *conf, char *key)
{
    char *ret;

    if (!conf) error("The config was called but not initialized.\n");

    ret = g_hash_table_lookup(conf, key);
    if ( !ret ) 
        error("Config: The variable \"%s\" not set.\n", key);

    return ret;
}


char *conf_get_string(GHashTable *vals, char *key, char *default_value)
{
    char *ret;
    
    if (!vals) return default_value;

    ret = g_hash_table_lookup(vals, key);
    if ( !ret )
        return default_value;
    return ret;
}


int conf_get_int_or_die(GHashTable *vals, char *key)
{
    char *v;

    if (!vals) error("The config was called but not initialized.\n");

    v = g_hash_table_lookup(vals, key);
    if ( ! v ) error("Config: The variable \"%s\" not set.\n", key);

    return atoi(v);
}


int conf_get_int(GHashTable *vals, char *key, int default_value)
{
    char *v;

    if (!vals) return default_value;

    v = g_hash_table_lookup(vals, key);
    if ( ! v ) 
        return default_value;

    return atoi(v);
}


float  conf_get_float_or_die(GHashTable *vals, char *key)
{
    char *v;

    if (!vals) error("The config was called but not initialized.\n");

    v = g_hash_table_lookup(vals, key);
    if ( ! v ) error("Config: The variable \"%s\" not set.\n", key);

    return atof(v);
}

float conf_get_float(GHashTable *vals, char *key, float default_value)
{
    char *v;

    if (!vals) return default_value;

    v = g_hash_table_lookup(vals, key);
    if ( ! v ) 
        return default_value;

    return atof(v);
}

double conf_get_double(GHashTable *vals, char *key, double default_value)
{
    char *v;

    if (!vals) return default_value;

    v = g_hash_table_lookup(vals, key);
    if ( ! v ) 
        return default_value;

    return atof(v);
}


int conf_is_string_in_list(GHashTable *vals, char *key, char *val)
{
    char *v;
    int len;

    if (!vals) return 0;

    v = g_hash_table_lookup(vals, key);
    if ( ! v ) return 0;

    len = strlen(val);

    while ( v && *v )
    {
        while ( *v && (isspace(*v) || *v==',') ) v++;
        if ( ! *v ) return 0;

        char *t = v;
        while ( *t && !isspace(*t) && *t!=',') t++;

        if ( len == t-v && !strncmp(val, v, t-v) ) return 1;
        v = t;
    }

    return 0;
}

int conf_is_int_in_list(GHashTable *conf, char *key, int value)
{
    char buf[MAX_LEN];
    snprintf(buf,MAX_LEN,"%d",value);
    return conf_is_string_in_list(conf,key,buf);
}

GList *conf_get_list_or_die(GHashTable *vals, char *key, char separator)
{
    GList *ret = conf_get_list(vals,key,separator,NULL);
    if ( !ret )
        error("Config: The variable \"%s\" not set.\n", key);
    return ret;
}


/* The separator character can be escaped by the backslash \. Do not
    use backslash as a separator.  */
GList *conf_get_list(GHashTable *vals, char *key, char separator, char *default_value)
{
    GList *ret = NULL;
    char *s,*str, *value, *remove;
    int n = 0;
    
    value = conf_get_string(vals,key,NULL);
    if ( value ) 
        value = strdup(value);
    else
    {
        if (!default_value) return NULL;
        value = strdup(default_value);
    }
    remove = value;
    
    s = value;
    while (s)
    {
        s = strchr(value, separator);
        if ( s>value && s[-1]=='\\' )
        {
            s++;
            continue;
        }

        if (s)
        {
            *s = 0;
            str = strndup(value,s-value);
        }
        else
            str = strdup(value);

        ret = g_list_append(ret,str);
        n++;

        if (!s)
            break;
        
        s++;
        value = s;
    }

    free(remove);
    return ret;
}

